#!/usr/bin/env python3
##############################################################################
# Copyright (c) 2022-2024
#
# Author(s):
#  Christian Hoffmann
#  The Jamulus Development Team
#
##############################################################################
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
##############################################################################

"""
Generates the JSON RPC documentation from the source code and writes it into
../docs/JSON-RPC.md.

Usage:
./tools/generate_json_rpc_docs.py

"""

import os
import re

source_files = [
    "src/rpcserver.cpp",
    "src/serverrpc.cpp",
    "src/clientrpc.cpp",
]

repo_root = os.path.join(os.path.dirname(__file__), '..')


class DocumentationItem:
    """
    Represents a documentation item. In the source code they look like this:

        /// @rpc_notification jamulusclient/channelLevelListReceived
        /// @brief Emitted when the channel level list is received.
        /// @param {array} params.channelLevelList - The channel level list.
        ///  Each item corresponds to the respective client retrieved from the
        ///  jamulusclient/clientListReceived notification.
        /// @param {number} params.channelLevelList[*] - The channel level,
        ///  an integer between 0 and 9.
    """

    def __init__(self, name, type_):
        """
        @param name: str
        @param type_: str
        """
        self.name = name
        self.type = type_
        self.brief = DocumentationText()
        self.params = []
        self.results = []
        self.current_tag = None

    def handle_tag(self, tag_name):
        """
        @param tag_name: str, brief|param|result
        """
        if tag_name == "brief":
            self.current_tag = self.brief
        elif tag_name == "param":
            self.current_tag = DocumentationText()
            self.params.append(self.current_tag)
        elif tag_name == "result":
            self.current_tag = DocumentationText()
            self.results.append(self.current_tag)
        else:
            raise ValueError("Unknown tag: " + tag_name)

    def handle_text(self, text):
        """
        @param text: str
        """
        self.current_tag.add_text(text)

    def sort_key(self):
        """
        @return str
        """
        return self.type + ": " + self.name

    def to_markdown(self):
        """
        @return: Markdown-formatted str with name, brief, params and results
        """
        output = ["### " + self.name, "", str(self.brief), ""]

        if len(self.params) > 0:
            output.append("Parameters:")
            output.append("")
            output.append(DocumentationTable(self.params).to_markdown())
            output.append("")

        if len(self.results) > 0:
            output.append("Results:")
            output.append("")
            output.append(DocumentationTable(self.results).to_markdown())
            output.append("")

        return "\n".join(output)


class DocumentationText:
    """
    Represents text inside the documentation.
    """

    def __init__(self):
        """Constructor"""
        self.parts = []

    def add_text(self, text):
        """
        @param text: str
        """
        self.parts.append(text)

    def __str__(self):
        """
        @return: Space-delimited list of previously added parts
        """
        return " ".join(self.parts)


class DocumentationTable:
    """
    Represents parameters and results table.
    """

    def __init__(self, tags):
        """
        @param tags: iterable
        """
        self.tags = tags

    def to_markdown(self):
        """
        @return: str containing the Markdown table
        """
        output = ["| Name | Type | Description |", "| --- | --- | --- |"]
        tag_re = re.compile(r"^{(\w+)}\s+(\S+)\s+-\s+(.*)$", re.DOTALL)
        for tag in self.tags:
            text = str(tag)
            # Parse tag in form of "{type} name - description"
            match = tag_re.match(text)
            if match:
                type_ = match.group(1)
                name = match.group(2)
                description = re.sub(r"^\s+", " ", match.group(3)).strip()
                output.append(f"| {name} | {type_} | {description} |")
        return "\n".join(output)


items = []
current_item = None

# Parse the source code
for source_file in source_files:
    with open(os.path.join(repo_root, source_file), "r") as f:
        for line in f.readlines():
            line = line.strip()
            if line.startswith("/// @rpc_notification "):
                current_item = DocumentationItem(
                    line[len("/// @rpc_notification "):], "notification"
                )
                items.append(current_item)
            elif line.startswith("/// @rpc_method "):
                current_item = DocumentationItem(
                    line[len("/// @rpc_method "):], "method"
                )
                items.append(current_item)
            elif line.startswith("/// @brief "):
                current_item.handle_tag("brief")
                current_item.handle_text(line[len("/// @brief "):])
            elif line.startswith("/// @param "):
                current_item.handle_tag("param")
                current_item.handle_text(line[len("/// @param "):])
            elif line.startswith("/// @result "):
                current_item.handle_tag("result")
                current_item.handle_text(line[len("/// @result "):])
            elif line.startswith("///"):
                current_item.handle_text(line[len("///"):])
            elif line == "":
                pass
            else:
                current_item = None

items.sort(key=lambda item: item.name)

PREAMBLE = """
# Jamulus JSON-RPC Server Documentation

<!--
This file is automatically generated from the source code.
Do not edit this file manually.
See `tools/generate_json_rpc_docs.py` for details.
-->

A JSON-RPC server can be added to both Jamulus client and Jamulus server to allow programmatic access.
To add the JSON-RPC server, run Jamulus with the `--jsonrpcport <port> --jsonrpcsecretfile /file/with/a/secret.txt` options.
This will start a JSON-RPC server on the specified port on the localhost.

The file referenced by `--jsonrpcsecretfile` must contain a single line with a freely chosen string with at least 16 characters.
It can be generated like this:
```
$ openssl rand -base64 10 > /file/with/a/secret.txt
```

The JSON-RPC server defaults to listening on the local loopback network interface (127.0.0.1).
This can be optionally changed by using the `--jsonrpcbindip <ip address>` command line option.


## Wire protocol

The JSON-RPC server is based on the [JSON-RPC 2.0](https://www.jsonrpc.org/specification) protocol, using [streaming newline-delimited JSON over TCP](https://clue.engineering/2018/introducing-reactphp-ndjson) as the transport. There are three main types of messages being exchanged:

- A **request** from the consumer to Jamulus.
- A **response** from Jamulus to the consumer.
- A **notification** from Jamulus to the consumer.

## Connect to a JSON-RPC server

On Linux, you can connect to a JSON-RPC server using the `nc` CLI tool. 

On Windows, [you can download](https://nmap.org/ncat/) and use the `ncat` CLI tool.

This snippet uses [jayson](https://www.npmjs.com/package/jayson) to connect using Node.js:

```
const jayson = require("jayson/promise");
const client = new jayson.client.tcp({ host: "127.0.0.1", port: 22100 });

client.request('jamulusserver/getServerInfo', {})
.then(console.log)
.catch(console.error)
```

## Example

After opening a TCP connection to the JSON-RPC server, the connection must be authenticated:

```json
{"id":1,"jsonrpc":"2.0","method":"jamulus/apiAuth","params":{"secret": "...the secret from the file in --jsonrpcsecretfile..."}}
```

Request must be sent as a single line of JSON-encoded data, followed by a newline character. Jamulus will send back a **response** in the same manner:

```json
{"id":1,"jsonrpc":"2.0","result":"ok"}
```
After successful authentication, the following **request** can be sent:

```json
{"id":2,"jsonrpc":"2.0","method":"jamulus/getMode","params":{}}
```

The request must be sent as a single line of JSON-encoded data, followed by a newline character. Jamulus will send back a **response** in the same manner:

```json
{"id":2,"jsonrpc":"2.0","result":{"mode":"client"}}
```

Jamulus will also send **notifications** to the consumer:

```json
{"jsonrpc":"2.0","method":"jamulusclient/chatTextReceived","params":{"text":"<font color=\\"mediumblue\\">(01:23:45 AM) <b>user</b></font> test"}}
```

"""

docs_path = os.path.join(repo_root, 'docs', 'JSON-RPC.md')
with open(docs_path, "w") as f:
    f.write(PREAMBLE)

    f.write("## Method reference\n")
    for item in filter(lambda item: item.type == "method", items):
        f.write(item.to_markdown() + "\n\n")

    f.write("## Notification reference\n")
    for item in filter(lambda item: item.type == "notification", items):
        f.write(item.to_markdown() + "\n\n")



#!/usr/bin/env python3
##############################################################################
# Copyright (c) 2022-2024
#
# Author(s):
#  Christian Hoffmann
#  The Jamulus Development Team
#
##############################################################################
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
##############################################################################

"""
This script is triggered from the GitHub Autobuild workflow.
It analyzes Jamulus.pro and git push details (tag vs. branch, etc.) to decide
  - whether a release should be created,
  - whether it is a pre-release, and
  - what its title should be.
"""

import os
import re
import subprocess

REPO_PATH = os.path.join(os.path.dirname(__file__), '..', '..')


def get_version_from_jamulus_pro():
    with open(REPO_PATH + '/Jamulus.pro', 'r') as f:
        pro_content = f.read()
    matches = re.search(r'^VERSION\s*=\s*(\S+)$', pro_content, re.MULTILINE)
    if not matches:
        raise ValueError("Unable to determine Jamulus.pro VERSION")
    return matches.group(1)


def get_git_hash():
    return subprocess.check_output([
        'git',
        'describe',
        '--match=xxxxxxxxxxxxxxxxxxxx',
        '--always',
        '--abbrev',
        '--dirty'
    ]).decode('ascii').strip()


def get_build_version(jamulus_pro_version):
    if "dev" in jamulus_pro_version:
        version = f"{jamulus_pro_version}-{get_git_hash()}"
        return 'intermediate', version

    version = jamulus_pro_version
    return 'release', version


def set_github_variable(varname, varval):
    print(f"{varname}='{varval}'")  # console output
    output_file = os.getenv('GITHUB_OUTPUT')
    with open(output_file, "a") as ghout:
        ghout.write(f"{varname}={varval}\n")

jamulus_pro_version = get_version_from_jamulus_pro()
set_github_variable("JAMULUS_PRO_VERSION", jamulus_pro_version)
build_type, build_version = get_build_version(jamulus_pro_version)
print(f'building a version of type "{build_type}": {build_version}')

full_ref = os.environ['GITHUB_REF']
publish_to_release = bool(re.match(r'^refs/tags/r\d+_\d+_\d+\S*$', full_ref))

# BUILD_VERSION is required for all builds including branch pushes
# and PRs:
set_github_variable("BUILD_VERSION", build_version)

# PUBLISH_TO_RELEASE is always required as the workflow decides about further
# steps based on this. It will only be true for tag pushes with a tag
# starting with "r".
set_github_variable("PUBLISH_TO_RELEASE", str(publish_to_release).lower())

if publish_to_release:
    ref_list = full_ref.split("/", 2)
    release_tag = ref_list[2]
    release_title = f"Release {build_version}  ({release_tag})"
    is_prerelease = not re.match(r'^r\d+_\d+_\d+$', release_tag)
    if not is_prerelease and build_version != release_tag[1:].replace('_', '.'):
        raise ValueError(f"non-pre-release tag {release_tag} doesn't match Jamulus.pro VERSION = {build_version}")

    # Those variables are only used when a release is created at all:
    set_github_variable("IS_PRERELEASE", str(is_prerelease).lower())
    set_github_variable("RELEASE_TITLE", release_title)
    set_github_variable("RELEASE_TAG", release_tag)



#!/usr/bin/python

from __future__ import print_function

from keras.models import Sequential
from keras.models import Model
from keras.layers import Input
from keras.layers import Dense
from keras.layers import LSTM
from keras.layers import GRU
from keras.models import load_model
from keras import backend as K
import sys

import numpy as np

def printVector(f, vector, name):
    v = np.reshape(vector, (-1));
    #print('static const float ', name, '[', len(v), '] = \n', file=f)
    f.write('static const opus_int8 {}[{}] = {{\n   '.format(name, len(v)))
    for i in range(0, len(v)):
        f.write('{}'.format(max(-128,min(127,int(round(128*v[i]))))))
        if (i!=len(v)-1):
            f.write(',')
        else:
            break;
        if (i%8==7):
            f.write("\n   ")
        else:
            f.write(" ")
    #print(v, file=f)
    f.write('\n};\n\n')
    return;

def binary_crossentrop2(y_true, y_pred):
        return K.mean(2*K.abs(y_true-0.5) * K.binary_crossentropy(y_pred, y_true), axis=-1)


#model = load_model(sys.argv[1], custom_objects={'binary_crossentrop2': binary_crossentrop2})
main_input = Input(shape=(None, 25), name='main_input')
x = Dense(32, activation='tanh')(main_input)
x = GRU(24, activation='tanh', recurrent_activation='sigmoid', return_sequences=True)(x)
x = Dense(2, activation='sigmoid')(x)
model = Model(inputs=main_input, outputs=x)
model.load_weights(sys.argv[1])

weights = model.get_weights()

f = open(sys.argv[2], 'w')

f.write('/*This file is automatically generated from a Keras model*/\n\n')
f.write('#ifdef HAVE_CONFIG_H\n#include "config.h"\n#endif\n\n#include "mlp.h"\n\n')

printVector(f, weights[0], 'layer0_weights')
printVector(f, weights[1], 'layer0_bias')
printVector(f, weights[2], 'layer1_weights')
printVector(f, weights[3], 'layer1_recur_weights')
printVector(f, weights[4], 'layer1_bias')
printVector(f, weights[5], 'layer2_weights')
printVector(f, weights[6], 'layer2_bias')

f.write('const DenseLayer layer0 = {\n   layer0_bias,\n   layer0_weights,\n   25, 32, 0\n};\n\n')
f.write('const GRULayer layer1 = {\n   layer1_bias,\n   layer1_weights,\n   layer1_recur_weights,\n   32, 24\n};\n\n')
f.write('const DenseLayer layer2 = {\n   layer2_bias,\n   layer2_weights,\n   24, 2, 1\n};\n\n')

f.close()



#!/usr/bin/env python3
"""
Generates a list of GitHub usernames who contributed to a specific release.

May need a GitHub Personal access token to avoid hitting rate limits.

Usage:

## Website repo (jamulussoftware/jamuluswebsite)
- cd jamuluswebsite
# Find commit ID of previous version. Search for "RELEASE:"
- git log --oneline release
- /path/to/jamulussoftware/code/tools/get_release_contributors.py \
    --from COMMIT_ID_FROM_PREVIOUS_STEP --to next-release \
    --github-token YOUR_TOKEN --repo .

## Code repo (jamulussoftware/jamulus)
- git pull -C /path/
- ./tools/get_release_contributors.py --from r3_8_1 --to r3_8_2 \
    --github-token YOUR_TOKEN --repo .

"""
import argparse
import logging
import os
import re
import subprocess

import requests
import yaml

logger = logging.getLogger('')


class UnexpectedGithubStatus(RuntimeError):
    """
    Exception which is raised when an API request fails.
    """


# List of contributor names which should be ignored (such as bots):
ignore_list = ['@github-actions[bot]', '@imgbot[bot]', '@actions-bot',
               '@actions-user', '@ImgBotApp', '@dependabot[bot]', '@weblate']

CHARSET = 'utf-8'


class Authors:
    """
    The Authors class provides methods to get the GitHub login of each
    commit in an efficient fashion.
    It uses the GitHub API and caches name+email-to-login associations
    to avoid excessive API usage.
    """

    def __init__(self, path):
        self.keys_to_user = {}
        self.path = path
        self.repo = None
        self.token = None
        try:
            self.load()
        except FileNotFoundError:
            pass

    def set_repo(self, repo):
        self.repo = repo

    def set_github_token(self, token):
        self.token = token

    def _get_login(self, key, commit_hash):
        """
        Returns the GitHub login associated with the given name+email key.
        A related commit hash is required in order to have a reference
        to use in the GitHub API request.

        Once looked up, the results are cached in a local file.
        """
        if key not in self.keys_to_user:
            user = self.get_user_by_email(key)
            if not user and commit_hash:
                user = self.get_user_by_commit(commit_hash)

            if user:
                self.keys_to_user[key] = user
                self.save()

        return self.keys_to_user.get(key, None)

    def get_login_or_realname(self, key, commit_hash):
        """
        Returns the Github login (@-prefixed) or, if not found, the real name from
        the given name+email key.
        """
        login = self._get_login(key, commit_hash)
        if login:
            return f'@{login}'
        matches = re.match(r'\A([^@<>]+) <.*>\Z', key)
        if matches:
            return matches.group(1)
        logger.warning("unable to extract GitHub login or real name from %r", key)
        return None

    def get_user_by_commit(self, sha):
        """
        Retrieves the associated GitHub username for the given commit hash.
        """
        resp = self._github_api_get(f'repos/jamulussoftware/{self.repo}/commits/{sha}')
        if 200 <= resp.status_code < 300:
            try:
                return resp.json()['author']['login']
            except (TypeError, KeyError):
                logger.warning('%s has not GitHub author, saving as empty', sha)
                return ''
        if resp.status_code == 422:
            logger.warning('unable to find author of %s, saving as empty', sha)
            return ''
        raise UnexpectedGithubStatus(f"status was {resp.status_code}")

    def _github_api_get(self, path, *args, **kwargs):
        headers = {
            'Accept': 'application/vnd.github.v3+json',
        }
        if self.token:
            headers['Authorization'] = f"token {self.token}"
        return requests.get(f'https://api.github.com/{path}', *args, headers=headers, **kwargs,
                            timeout=10)

    def get_user_by_email(self, key):
        m = re.match(r'\A[^<]+<([^<> ]+@[^<> ]+)>\Z', key)
        if not m:
            return None
        email = m.group(1)
        # Handle Github-generated email addresses via static matching:
        m = re.match(r'\A(\d+\+)?([^+@]+)@users\.noreply\.github\.com\Z', email)
        if m:
            return m.group(2)
        resp = self._github_api_get('search/users', params={'q': f'{email} in:email'})
        if resp.status_code < 200 or resp.status_code >= 300:
            logger.warning('search/users for %s failed with code %s', email, resp.status_code)
            return None
        items = resp.json().get('items', [])
        for item in items:
            login = item['login']
            u = self._github_api_get(f'users/{login}').json()
            if u.get('email', '') == email:
                return login

        logger.warning('unable to find a github profile with public email %s', email)
        return None

    def save(self):
        """
        Saves the cache to disk.
        """
        with open(self.path, 'w') as f:
            yaml.dump(self.keys_to_user, f)

    def load(self):
        """
        Loads the cache from disk.
        """
        with open(self.path, 'r') as f:
            self.keys_to_user = yaml.safe_load(f) or {}


# Store the lookup cache right next to this script:
authors_cache = __file__.replace('.py', '-cache.yaml')

# Prepare a global Authors instance
authors = Authors(authors_cache)


def main(from_, to):
    if os.path.exists('src'):
        logger.info('scanning code repo')
        authors.set_repo('jamulus')
        print_code_contributors(from_, to)
        print_app_translators(from_, to)
    elif os.path.exists('wiki'):
        authors.set_repo('jamuluswebsite')
        print_website_contributors(from_, to)
        logger.info('scanning website repo')
    else:
        logger.error('is this a code or website repo?')


def print_code_contributors(from_, to):
    return print_contributors('Code contributors', ['.', ':!src/translation'], from_, to)


def print_app_translators(from_, to):
    return print_contributors('Application translators', ['src/translation'], from_, to)


def print_website_contributors(from_, to):
    return print_contributors('Website contributors/translators', ['.'], from_, to)


def print_contributors(title, git_log_selector, from_, to):
    contributors = [u for u in find_contributors(git_log_selector, from_, to) if
                    u and u not in ignore_list]
    contributors_str = ', '.join(contributors)
    print(f'{title}: {contributors_str}')


def find_contributors(git_log_selector, from_ref, to_ref):
    """
    Uses `git log` with the provided git_log_selector to list all commits
    starting with git reference `from_ref` and going until reference `to_ref`.

    A `git_log_selector` could just be a path such as '.'.
    `from_` and `to` can be any committish such as a commit hash or a tag.
    """
    contributors = set()
    co_author_keys = set()
    commits = subprocess.check_output(
        ['git', 'log', '-z', '--show-pulls', '--format=format:%H %an <%ae>%n%b',
         f'{from_ref}..{to_ref}', '--'] + git_log_selector)
    commits = commits.decode(CHARSET)
    for commit in commits.split('\0'):
        if not commit:
            continue
        sha, author_key = commit.split('\n', 1)[0].split(' ', 1)
        contributor = authors.get_login_or_realname(author_key, sha)
        contributors.add(contributor)
        co_authors = re.findall(r'Co-authored-by:\s*(\S.*(<[^ >]+>))\s*(?:$|\n)', commit, re.I)
        for co_author_full, co_author_email in co_authors:
            logger.debug('checking co author %s', co_author_full)
            contributor = authors.get_login_or_realname(co_author_full, None)
            if not contributor or not contributor.startswith('@'):
                # try to find a previous commit by this mail address
                # and pass this commit id to get_login_or_realname() to retrieve the
                # associated handle from the GitHub API.
                esc_email = re.escape(co_author_email)
                commit = subprocess.check_output(['git', 'log', '--format=%H', '--max-count=1',
                                                  f'--author={esc_email}']).strip().decode(CHARSET)
                if commit:
                    contributor = authors.get_login_or_realname(co_author_full, commit)
            if contributor:
                contributors.add(contributor)

    # Resolve co-authors last because we have to rely on having seen the
    # email-to-login mapping via some other commit.
    for co_author in co_author_keys:
        contributor = authors.get_login_or_realname(co_author, None)
        if contributor:
            contributors.add(contributor)
        else:
            # TODO: check this is the right logic...
            contributors.add(co_author)
    return sorted(contributors, key=str.casefold)


if __name__ == '__main__':
    p = argparse.ArgumentParser(
        description='Generates a list of Github user names who contributed to a specific release.')
    p.add_argument('--from', dest='from_', required=True,
                   help='the first git hash or tag to include in the analysis, '
                        'e.g. the tag of the previous release')
    p.add_argument('--to', required=True,
                   help='the last git hash or tag to include in the analysis, '
                        'e.g. the (rc) tag of the target release')
    p.add_argument('--repo', required=True,
                   help='the path to the git repository to be analyzed, e.g. ./jamuluswebsite')
    p.add_argument('--github-token',
                   help='a Github Personal Access Token; optional, but might be needed '
                        'if we exceed the anonymous API requests per hour limit')
    p.add_argument('--verbose', '-v', action='store_true',
                   help='enable verbose output')
    p.add_argument('--quiet', '-q', action='store_true',
                   help='only log errors')
    args = p.parse_args()
    if args.verbose and args.quiet:
        p.error('--verbose and --quiet are mutually exclusive')
    if args.verbose:
        level = logging.DEBUG
    elif args.quiet:
        level = logging.ERROR
    else:
        level = logging.WARNING
    logging.basicConfig(format='%(levelname)s %(message)s', level=level)
    os.chdir(args.repo)
    authors.set_github_token(args.github_token)
    main(args.from_, args.to)



#!/usr/bin/env python3
##############################################################################
# Copyright (c) 2022-2024
#
# Author(s):
#  Christian Hoffmann
#  The Jamulus Development Team
#
##############################################################################
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
#
##############################################################################

"""
Jamulus uses Qt5 QLocale::Country codes in its protocol.
Qt6 changed those codes in an incompatible way.
To make >=Qt6 clients compatible with the Jamulus protocol in the wild,
we need to translate between those codes.
This script generates two translation tables from Qt5 + Qt6 header files:
A table to translate from Qt5 to Qt6 and an inverse table.
"""
import re

# verify that the following actually points to Qt5 on your system!
QT5_INCLUDE = '/usr/include/qt/QtCore/qlocale.h'
QT6_INCLUDE = '/usr/include/qt6/QtCore/qlocale.h'


def parse_enum_from_header(header):
    name_to_val = {}
    val_to_name = {}
    with open(header) as f:
        s = f.read()
    matches = re.search(r'enum Country[^\n]+{([^}]+)}', s)
    assignments = matches.group(1)

    assign_pattern = re.compile(r'^\s*(\S+)(?:\s+.*)?\s+=\s+([^\s,]+),?$')
    for assignment in assignments.split('\n'):
        if assignment.startswith('#') or not assignment.strip():
            continue
        matches = assign_pattern.match(assignment)
        if not matches:
            raise ValueError(f"unable to match line: {assignment!r}")
        name = matches.group(1)
        val = matches.group(2)
        if val.isnumeric():
            val = int(val)
            if val not in val_to_name:
                val_to_name[val] = name
            name_to_val[name] = val
        elif val in name_to_val:
            # resolve aliases:
            name_to_val[name] = name_to_val[val]
        else:
            raise ValueError("Unhandled case")
    return name_to_val, val_to_name


def make_struct(name, table):
    ret = ''
    highest_key = max(int(x) for x in table.keys())
    # Need to escape the literal '{' by using '{{'
    ret += f'constexpr int const static {name}[] = {{'
    for key in range(highest_key + 1):
        ret += str(table.get(key, -1))  # fill gaps with -1
        ret += ', '
    ret += '};'
    return ret


qt5_name_to_val, qt5_val_to_name = parse_enum_from_header(QT5_INCLUDE)
qt6_name_to_val, qt6_val_to_name = parse_enum_from_header(QT6_INCLUDE)

qt5_to_qt6 = {}
qt6_to_qt5 = {}
# We need to support all Qt5 codes, so we work by value:
for qt5_val, name in qt5_val_to_name.items():
    qt6_val = qt6_name_to_val[name]
    qt5_to_qt6[qt5_val] = qt6_val
    qt6_to_qt5[qt6_val] = qt5_val

print(make_struct('wireFormatToQt6Table', qt5_to_qt6))
print(make_struct('qt6CountryToWireFormat', qt6_to_qt5))
